
""" KnackHQ REST Client. """


import collections
import json
import os

import certifi
import urllib3
from .knackhq import KnackHQObject


class KnackHQClient(collections.Iterable):
    """ KnackHQ root client.

        Use arguments or ENV variables to initialize client.

        Optional ENV variables are:
            * KNACKHQ_APP_ID
            * KNACKHQ_API_KEY
            * KNACKHQ_ENDPOINT

        Arguments:
            app_id   (str):  Application ID string
            api_key  (str):  API key
            endpoint (str):  KnackHQ endpoint (default: https://api.knackhq.com/v1)
    """
    def __init__(self, app_id=None, api_key=None, endpoint=None):
        self._app_id = app_id or os.getenv('KNACKHQ_APP_ID')
        self._api_key = api_key or os.getenv('KNACKHQ_API_KEY')
        self._endpoint = endpoint \
            or os.getenv('KNACKHQ_ENDPOINT') \
            or 'https://api.knackhq.com/v1'

    def __repr__(self):
        return "<KnackHQClient %s>" % self._endpoint

    def __iter__(self):
        endpoint = os.path.join(self._endpoint, 'objects')
        response = self.request(endpoint)
        for item in response.get('objects', []):
            yield self.get_object(item['key'])

    def __len__(self):
        endpoint = os.path.join(self._endpoint, 'objects')
        response = self.request(endpoint)
        return len(response.get('objects', []))

    def request(self, endpoint, verb='GET', **kwargs):
        """ Send a request to the API.

            Arguments:
                endpoint (str):   URL for request
                verb     (str):   Optional verb of request
                kwargs   (dict):  Optional additional request arguments

            Returns:
                JSON response of request.

            Raises:
                ValueError on bad response.
        """
        head = {
            "Content-Type": "application/json",
            "X-Knack-Application-Id": self._app_id,
            "X-Knack-REST-API-Key": self._api_key}
        http = urllib3.PoolManager(cert_reqs='CERT_REQUIRED', ca_certs=certifi.where())
        resp = http.request(verb, endpoint, headers=head, **kwargs)
        try:
            return json.loads(resp.data)
        except ValueError as err:
            raise ResponseError(err + "\n" + resp.data)

    def add_object(self, name):
        """ Add an object to the app.

            Arguments:
                name (str):  Name of new object

            Returns:
                KnackHQObject instance

            Raises:
                DuplicateObjectError if object with same name already exists
        """
        try:
            self._object_key(name)
        except ObjectNotFoundError:
            endpoint = os.path.join(self._endpoint, 'objects')
            body = json.dumps({'name': name})
            response = self.request(endpoint, 'POST', body=body)
            endpoint = os.path.join(self._endpoint, 'objects', response['object']['key'])
            return KnackHQObject(self, endpoint, **response)

        raise DuplicateObjectError("More than one object named '%s'" % name)

    def get_object(self, key_or_name):
        """ Get object by key or name. If a key is not provided it
            will be found by filtering on all objects. It is more
            efficient to use the key for getting objects.

            Arguments:
                key_or_name (str):  Object key or name

            Returns:
                KnackHQObject instance.

            Raises:
                DuplicateObjectError on multiple objects with the same name.
                ObjectNotFoundError on object not found.
        """

        try:
            endpoint = os.path.join(self._endpoint, 'objects', key_or_name)
            response = self.request(endpoint)
        except ResponseError:
            endpoint = os.path.join(self._endpoint, 'objects', self._object_key(key_or_name))
            response = self.request(endpoint)

        return KnackHQObject(client=self, endpoint=endpoint, **response)

    def _object_key(self, name):
        """ Helper to get object key from object name.

            Arguments:
                name (str):  Name of object

            Returns:
                Key of object.

            Raises:
                DuplicateObjectError on multiple objects with the same name.
                ObjectNotFoundError on object not found.
        """
        objects = [x for x in self if x['name'] == name]
        if len(objects) == 1:
            return objects[0]['key']
        elif len(objects) > 1:
            raise DuplicateObjectError("More than one object named '%s'" % name)
        raise ObjectNotFoundError(name)


class ResponseError(ValueError):
    """ KnackHQ Client Error. """
    pass


class DuplicateObjectError(ValueError):
    """ KnackHQ Client Error. """
    pass


class ObjectNotFoundError(KeyError):
    """ KnackHQ Client Error. """
    pass


class DuplicateFieldError(ValueError):
    """ KnackHQ Client Error. """
    pass



""" KnackHQ Objects/Records """


import collections
import json
import os
import re

# Python2
try:
    from urllib import quote_plus
    from urlparse import urlparse

# Python3
except ImportError:
    # pylint: disable=ungrouped-imports
    from urllib.parse import urlparse
    from urllib.parse import quote_plus


class KnackHQRecord(collections.Mapping):
    """ KnackHQ Record.
    """
    @property
    def object(self):
        """ Memoized response of a GET on the record parent. """
        if self._object is None:
            endpoint = re.split(r'/records/.*?', self._endpoint)[0]
            self._object = KnackHQObject(self._client, endpoint)
        return self._object

    @property
    def record(self):
        """ Memoized response of a GET on the record. """
        if self._record is None:
            self._record = self._client.request(self._endpoint)
        return self._record

    def __init__(self, client=None, endpoint=None, record=None):
        self._client = client
        self._endpoint = endpoint
        self._object = None
        self._record = record

    def __repr__(self):
        return "<KnackHQRecord %s>" % urlparse(self._endpoint).path

    def __getitem__(self, key):
        try:
            return self.record[key]
        except KeyError:
            fields = [x for x in self.object['fields'] if x['name'] == key]
            if len(fields) == 1:
                field_key = fields[0]['key']
                field_raw = "%s_raw" % field_key
                is_field = lambda x: x in (field_key, field_raw)
                return dict([(key, val) for key, val in self.record.items() if is_field(key)])

            raise KeyError("More than one field named '%s'" % key)

    def __iter__(self):
        return iter(self.record)

    def __len__(self):
        return len(self.record)


class KnackHQObject(collections.Iterable):
    """ KnackHQ Object (ie, table).

        Arguments:
            client   (Client):  KnackHQ Client instance
            endpoint (str):     URL to object resource

        Iterate over records in an object with:

            for record in object:
                # do something

        Iterate over filtered records in an object with:

            filters = [
                {
                    'field':    'field_1',
                    'operator': 'is',
                    'value':    'test'
                }, {
                    'field':    'field_2',
                    'operator': 'is not blank'
                }
            ]

            for record in object.where(filters=filters):
                # do something
    """
    @property
    def object(self):
        """ Memoized response of a GET on the object. """
        if self._object is None:
            self._object = self._client.request(self._endpoint)['object']
        return self._object

     # pylint: disable=redefined-builtin
    def __init__(self, client=None, endpoint=None, object=None):
        self._client = client
        self._endpoint = endpoint
        self._object = object  # pylint: disable=redefined-builtin

    def __repr__(self):
        return "<KnackHQObject %s>" % urlparse(self._endpoint).path

    def __getitem__(self, key):
        return self.object[key]

    def __iter__(self):
        return self.where()

    def __len__(self):
        endpoint = os.path.join(self._endpoint, 'records')
        try:
            page = self._client.request(endpoint)
            return page['total_records']
        except ValueError:
            return 0

    def keys(self):
        """ Return keys of KnackHQ object definiton. """
        return self.object.keys()

    def where(self, **kwargs):
        """ Get records from a REST request.

            Iterate over filtered records in an object with:

                filters = [
                    {
                        'field':    'field_1',
                        'operator': 'is',
                        'value':    'test'
                    }, {
                        'field':    'field_2',
                        'operator': 'is not blank'
                    }
                ]

                for record in object.where(filters=filters):
                    # do something

            Arguments:
                record_id     (str):   Optional ID of record
                page          (int):   Optional page number
                rows_per_page (int):   Optional rows per page number
                sort_field    (str):   Optional sort field key
                sort_order    (str):   Optional sort order
                filters       (list):  Optional list of filter dicts

            Returns:
                REST response JSON dict.
        """
        # Yield single record
        if 'record_id' in kwargs:
            record_id = kwargs['record_id']
            endpoint = os.path.join(self._endpoint, 'records', record_id)
            yield KnackHQRecord(self._client, endpoint)

        # Yield records
        else:
            endpoint = os.path.join(self._endpoint, 'records')
            endpoint += '?'
            for key, val in kwargs.items():
                if key == 'filters':
                    val = quote_plus(json.dumps(val))
                endpoint += "%s=%s&" % (key, val)

            # Yield records
            try:
                page = self._client.request(endpoint)
            except ValueError:
                raise StopIteration
            for record in page['records']:
                endpoint = os.path.join(self._endpoint, 'records', record['id'])
                yield KnackHQRecord(self._client, endpoint, record)

            # Recurse
            kwargs['page'] = int(page['current_page']) + 1
            self.where(**kwargs)
